from . import raw_nodes


def filter_resource_description(raw_rd: raw_nodes.Model, weights_priority_order=None) -> raw_nodes.Model:
    raise NotImplementedError("Packaging a model RDF of format version 0.1. is not supported.")
